//
//
//    guitune - program for tuning instruments (actually an oscilloscope)
//    Copyright (C) 1999  Florian Berger
//    Email: florian.berger@jk.uni-linz.ac.at
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License Version 2 as
//    published by the Free Software Foundation;
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//

// 22-08-2003: Added ALSA support, and included config.h and added #ifdefs for audio system selection at compile time - Reinier Lamers
// 29-03-2003: changed font from 'System' to QApplication::font()
//             switched to new audio input system
//             added getDSPName(), getSoundInputType(), setSoundInputType() methods
// Modified 2002 by Reinier Lamers

//#include <iostream.h>
//#include <sys/types.h>
//#include <sys/stat.h>
//#include <fcntl.h>
//#include <sys/ioctl.h>
//#include <linux/soundcard.h>

#include <config.h>

#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>


#include <qapplication.h>
#include <qwidget.h>
#include <qpainter.h>
#include <qfile.h>
#include <qfiledialog.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qscrollbar.h>
#include <qkeycode.h>
#include <qlineedit.h>
#include <qpushbutton.h>
#include <qimage.h>
#include <qtimer.h>
#include <qlcdnumber.h>
#include <qspinbox.h>
#include <qmainwindow.h>
#include <kmessagebox.h>
#include <klocale.h>


#include <math.h>

#include "osziview.h"
#include "logview.h"
#include "soundinput.h"
#include "guitune.h"
#include "audiodlg.h"

#include "resources.h"



//globally
double KAMMERTON, KAMMERTON_LOG;

void MainWidget::setTuningNorm()
{
    KAMMERTON=KAMMERTON_NORM;
    KAMMERTON_LOG=KAMMERTON_LOG_NORM;
    update();
}

void MainWidget::setTuningWien()
{
    KAMMERTON=KAMMERTON_WIEN;
    KAMMERTON_LOG=KAMMERTON_LOG_WIEN;
    update();
}

void MainWidget::setTuningPhys()
{
    KAMMERTON=KAMMERTON_PHYS;
    KAMMERTON_LOG=KAMMERTON_LOG_PHYS;
    update();
}


void MainWidget::setTuningEqui()
{
    logview->nat_tuning_on=false;
    update();
}


void MainWidget::setTuningNat()
{
    logview->nat_tuning_on=true;
    update();
}


void MainWidget::setScaleUS()
{
    logview->setScale(LogView::us_scale);
}

void MainWidget::setScaleUSAlt()
{
    logview->setScale(LogView::us_scale_alt);
}

void MainWidget::setScaleGE()
{
    logview->setScale(LogView::german_scale);
}

void MainWidget::setScaleGEAlt()
{
    logview->setScale(LogView::german_scale_alt);
}

void MainWidget::setScaleFR()
{
    logview->setScale(LogView::french_scale);
}

void MainWidget::setScaleFRAlt()
{
    logview->setScale(LogView::french_scale_alt);
}

void MainWidget::showLogView()
{
 	  logview->show();
    resizeEvent(0);   // rearrange widgets
}

void MainWidget::hideLogView()
{
 	  logview->hide();
    resizeEvent(0);
}

void MainWidget::showOszi()
{
 	  oszi->show();
    resizeEvent(0);
}

void MainWidget::hideOszi()
{
 	  oszi->hide();
    resizeEvent(0);
}

void MainWidget::setSampFreq(int f)
{
    sampfreq = f;

    must_restart_audio = true;
}


void MainWidget::setSampNr(int nr)
{
    sampnr=nr;
    oszi->setSampleNr(nr);
    emit signalSampNrChanged();
}

/* Sets whether notch filter is done or not, depending on the state of the
   notch checkbox.
*/
void MainWidget::setNotch() {
    oszi->setNotch(notch_input->isChecked());
}

/* These functions can't be entirely ifdeffed away because they're slots, and moc doesn't like ifdeffed-out slots */
void MainWidget::setDSPName(const char *name)
{
#ifdef USE_OSS
  strncpy(dsp_devicename,name, sizeof(dsp_devicename)-1);
  dsp_devicename[sizeof(dsp_devicename)-1] = '\0';

  must_restart_audio = true;
#endif
}

void MainWidget::setALSAName( const char *name )
{
#ifdef USE_ALSA
  strncpy( alsa_devicename, name, sizeof(alsa_devicename) - 1 );
  alsa_devicename[sizeof(alsa_devicename)-1] = '\0';

  must_restart_audio = true;
#endif
}

void MainWidget::showAudioDialog( void )
{
  AudioDialog *audiodialog = new AudioDialog( this, this );
  audiodialog->exec();
  delete audiodialog;

  if (must_restart_audio)
  {
    stop_audio();
    start_audio();
    must_restart_audio = false;
  }
}

/*
 * Following getSomething() methods are for now meant for the K2GuituneApp::saveOptions() thing
 */
#ifdef USE_OSS
char *MainWidget::getDSPName( void )
{
  static char ret[sizeof(dsp_devicename)];

  strcpy(ret, dsp_devicename);
  return ret;
}
#endif

#ifdef USE_ALSA
char *MainWidget::getALSAName( void )
{
  static char ret[sizeof(alsa_devicename)];

  strcpy( ret, alsa_devicename );
  return ret;
}
#endif

MainWidget::SoundInputType MainWidget::getSoundInputType( void )
{
  return soundinputtype;
}

void MainWidget::setSoundInputType( MainWidget::SoundInputType _soundinputtype )
{
  if (_soundinputtype != soundinputtype)
  {
    soundinputtype = _soundinputtype;
    must_restart_audio=true;
  }
}

MainWidget::MainWidget( QWidget *parent, const char *name, int argc, char **argv )
               : QWidget(parent,name)
{int i;
   //InitAudio();
   setMinimumSize(490,240);

   KAMMERTON=KAMMERTON_NORM;
   KAMMERTON_LOG=KAMMERTON_LOG_NORM;

   strcpy(dsp_devicename,"/dev/dsp");

   blksize=256; //might be a nice default, dunno :)
   sampnr=2048; // 
   sampfreq=44100; // Best to be as high as possible!
   sampfreq_exact = sampfreq;
   processing_audio=0;

   /* No longer initialize audio here. It happens automatically when K2GuituneApp::readOptions()
    * calls start_audio(). Initialize soundinput to 0 so we know we have nothing */
   //start_audio();
   soundinput = 0;

   freqs[0]=KAMMERTON; lfreqs[0]=KAMMERTON_LOG;
   for(i=1;i<12;i++){
      freqs [i] = freqs [i-1] * D_NOTE;
      lfreqs[i] = lfreqs[i-1] + D_NOTE_LOG;
   }

   oszi = new OsziView( this, "oszi_name" );
//   oszi->setGeometry(0,0,400,150);
   oszi->setSampleNr( sampnr );
   //oszi->setSampleFreq(sampfreq_exact); <-- gets done in start_audio() now
   logview = new LogView( this, "logview_name" );
   logview->setFrameStyle( QFrame::Box | QFrame::Sunken );
   logview->setLineWidth( 2 );
//   logview->setGeometry(0,oszi->height(),width(),80);  // watchow
   QColorGroup cg(qRgb(0xFF,0,0),qRgb(0,0,0),qRgb(0xFF,0x88,0x88),qRgb(0x88,0,0),qRgb(0xFF,0,0),qRgb(0xFF,0xFF,0xFF),qRgb(0x88,0x88,0x88));
   QPalette pal(cg,cg,cg);
   freqview = new QLCDNumber( this, "freqview_name");
   freqview->setNumDigits(9);
   freqview->setFrameStyle(QFrame::NoFrame);
   freqview->setSegmentStyle(QLCDNumber::Filled);
   freqview->setPalette(pal);
   nfreqview = new QLCDNumber( this, "nfreqview_name");
   nfreqview->setNumDigits(9);
   nfreqview->setFrameStyle(QFrame::NoFrame);
   nfreqview->setSegmentStyle(QLCDNumber::Filled);
   nfreqview->setPalette(pal);
//   freqview->setSegmentStyle(QLCDNumber::Filled);
   sampfreq_input = new QSpinBox( 1000, 48000, 50, this );
//   sampfreq_input -> setGeometry( freqview->x()+freqview->width()+30,
//                                  logview->y()+logview->height()+20,
//                                  70,20 );
   connect( sampfreq_input, SIGNAL(valueChanged(int)),
            this,           SLOT(setSampFreq(int))     );
   sampnr_input = new QSpinBox( 32, 11048, 32, this );
//   sampnr_input -> setGeometry( sampfreq_input->x(),
//                                sampfreq_input->y()+
//                                sampfreq_input->height()+10,  70,20 );
   sampnr_input -> setValue(sampnr);
   connect( sampnr_input, SIGNAL(valueChanged(int)),
            this,         SLOT(setSampNr(int))   );
   notch_input = new QCheckBox(this);
   connect( notch_input, SIGNAL(toggled(bool)),
            this,        SLOT(setNotch())  );
   notch_input->setChecked(false);
   timer = new QTimer(this);
   connect(  timer, SIGNAL(timeout()), this, SLOT(proc_audio()) );
}

void MainWidget::start_audio( void )
{
  /* the numerical arguments to the SoundInput::* functions are passed by
   * reference so they get set to the right values */
  switch( soundinputtype )
  {
  case autodetect:
    soundinput = SoundInput::autodetect(alsa_devicename, dsp_devicename, blksize, sampfreq, sampfreq_exact);
    break;
#ifdef USE_OSS
  case OSS:
    try {
      soundinput = new OSSSoundInput(dsp_devicename, blksize, sampfreq, sampfreq_exact);
    } catch (SoundInputException) {
      soundinput = 0;
    }
    break;
#endif
#ifdef USE_ARTS
  case aRts:
    try {
      soundinput = new ArtsSoundInput(blksize, sampfreq, sampfreq_exact);
    } catch (SoundInputException) {
      soundinput = 0;
    }
    break;
#endif
#ifdef USE_ALSA
  case ALSA:
    try {
      soundinput = new ALSASoundInput( alsa_devicename, blksize, sampfreq, sampfreq_exact );
    } catch (SoundInputException) {
      soundinput = 0;
    }
    break;
#endif
#ifdef USE_JACK
  case JACK:
    try {
      soundinput = new JACKSoundInput(blksize, sampfreq, sampfreq_exact);
    } catch (SoundInputException) {
      soundinput = 0;
    }
#endif
  }

  sampfreq_input->blockSignals( true );
  sampfreq_input->setValue(sampfreq);
  sampfreq_input->blockSignals( false );
  oszi->setSampleFreq(sampfreq_exact);
  emit signalSampFreqChanged();

  if (!soundinput)
  {
    KMessageBox::sorry( this, i18n("K3Guitune failed to initialize any audio system and thus cannot get any sound input. Check your audio settings (via Options->Audio Options... in the menubar)."), i18n("No audio system found"));
  } else {
    printf( "Audio driver initialized\n" );
    timer->start(0);
  }
}

void MainWidget::stop_audio( void )
{
  if (soundinput)
  {
    delete soundinput;
  }
  timer->stop();
}


void MainWidget::proc_audio()
{
  /* as of now, must_restart_audio is never true when proc_audio() is called. When once a ghost comes in and
   * starts calling MainWidget::set* things in the middle of our operation, this should be here though */
  #if 0
  if (must_restart_audio)
  {
    stop_audio();
    start_audio();
    must_restart_audio = false;
  }
  #endif

#if 0
 int i,j,n,trig,trigpos;
 static int k=0;
 unsigned char *c;
 double ldf,mldf;
 char str[50];
   processing_audio=1;
   trigpos=0;
   c=sample; n=0;
   n = soundinput->getData(c, blksize);
   for( i=0; i<n && Abs(c[i]-128)<2; i++ );
   //i=-1;
   j=0; trig=0;
   if (i<n) do {
      for( ; i<n-1; i++ )   /* n-1 because of POSTRIG uses i+1 */
          if ( POSTRIG(c,i) ) { trig=1; trigpos=i; }
      if( trig==0 ){
          n = soundinput->getData(c, blksize);
          j++;
          i=0;
      }
   }while( (!trig) && j<100 );
//   else printf("No Signal %d\n",blksize);
   if( trig ){
      for( i=n-trigpos; i<sampnr; i+=n ){
         c+=n;
         n = soundinput->getData(c, blksize);
      }
//      oszi->setSampleFreq( sampfreq_exact );
      oszi->setSamplePtr( &sample[trigpos] );
//      oszi->setSampleNr( sampnr );
      oszi->paintSample();
      freq_0t = (double)sampfreq*oszi->getfreq_s(1, 0, 1);
      lfreq_0t = log(freq_0t);
      while ( lfreq_0t < lfreqs[0]-D_NOTE_LOG/2.0 ) lfreq_0t+=LOG_2;
      while ( lfreq_0t >= lfreqs[0]+LOG_2-D_NOTE_LOG/2.0 ) lfreq_0t-=LOG_2;
      mldf=D_NOTE_LOG; note_0t=0;
      for( i=0; i<12; i++ ){
				 ldf = fabs(lfreq_0t-lfreqs[i]);
				 if (ldf<mldf) { mldf=ldf; note_0t=i; }
      }
      logview->change_lfreq(lfreq_0t);
      sprintf(str,"%.3f",freq_0t);
      freqview->display(str);
      double nfreq_0t=freqs[note_0t];
      while( nfreq_0t/freq_0t > D_NOTE_SQRT ) nfreq_0t/=2.0;
      while( freq_0t/nfreq_0t > D_NOTE_SQRT ) nfreq_0t*=2.0;
      sprintf(str,"%.3f",nfreq_0t);
      nfreqview->display(str);
//      printf("Note: %s (%lfHz) Freq=%lf\n",
//	        note[note_0t],freqs[note_0t],freq_0t);
   }
   k++;
   processing_audio=0;
#endif
 int i,j,n,trig,trigpos;
 static int k=0;
 short int *s;
 short int *sample_s16le;
 double ldf,mldf;
 char str[50];

   processing_audio=1;
   trigpos=0;
   sample_s16le = (short int *)sample;
   s = sample_s16le;
   n=0;

   n = soundinput->getData(s, blksize);

   for( i=0; i<n && Abs(s[i])<256; i++ );

   //i=-1;
   j=0; trig=0;
   if (i<n) do {
      for( ; i<n-1; i++ )   /* n-1 because of POSTRIG uses i+1 */
          if ( POSTRIG_S16(s,i) ) { trig=1; trigpos=i; }
      if( trig==0 ){
          n = soundinput->getData(s, blksize);
          j++;
          i=0;
      }
   }while( (!trig) && j<100 );
//   else printf("No Signal %d\n",blksize);
   if( trig ){
      for( i=n-trigpos; i<sampnr; i+=n ){
          s+=n;
          n = soundinput->getData(s, blksize);
      }

      oszi->setSamplePtr( &sample_s16le[trigpos] );

//      oszi->setSampleNr( sampnr );
      oszi->paintSample();
      freq_0t = (double)sampfreq*oszi->getfreq_s();
      lfreq_0t = log(freq_0t);
      while ( lfreq_0t < lfreqs[0]-D_NOTE_LOG/2.0 ) lfreq_0t+=LOG_2;
      while ( lfreq_0t >= lfreqs[0]+LOG_2-D_NOTE_LOG/2.0 ) lfreq_0t-=LOG_2;
      mldf=D_NOTE_LOG; note_0t=0;
      for( i=0; i<12; i++ ){
          ldf = fabs(lfreq_0t-lfreqs[i]);
          if (ldf<mldf) { mldf=ldf; note_0t=i; }
      }
      logview->change_lfreq(lfreq_0t);
      sprintf(str,"%0.3f",freq_0t);
      freqview->display(str);
      double nfreq_0t=freqs[note_0t];
      while( nfreq_0t/freq_0t > D_NOTE_SQRT ) nfreq_0t/=2.0;
      while( freq_0t/nfreq_0t > D_NOTE_SQRT ) nfreq_0t*=2.0;
      sprintf(str,"%0.3f",nfreq_0t);
      nfreqview->display(str);

//      printf("Note: %s (%lfHz) Freq=%lf\n",
//              note[note_0t],freqs[note_0t],freq_0t);
   }
   k++;
   processing_audio=0;
}



void MainWidget::paintEvent( QPaintEvent * )
{
	 QPainter p(this);

     p.setFont(QFont(QApplication::font().family(), 8));
     p.setPen( qRgb(0xFF,0xFF,0xFF) );
	   p.drawText( freqview->x()-p.fontMetrics().width(i18n("Freq.:"))-5+1,
	  	           freqview->y()+p.fontMetrics().ascent()+1
			           +p.fontMetrics().height(), i18n("Freq.:") );
     p.setPen( qRgb(0,0,0) );
	   p.drawText( freqview->x()-p.fontMetrics().width(i18n("Freq.:"))-5,
	  	           freqview->y()+p.fontMetrics().ascent()
			           +p.fontMetrics().height(), i18n("Freq.:") );

     p.setPen( qRgb(0xFF,0xFF,0xFF) );
	   p.drawText( nfreqview->x()-p.fontMetrics().width(i18n("Note:"))-5+1,
	  	           nfreqview->y()+p.fontMetrics().ascent()+1
			           +p.fontMetrics().height(), i18n("Note:") );
     p.setPen( qRgb(0,0,0) );
	   p.drawText( nfreqview->x()-p.fontMetrics().width(i18n("Note:"))-5,
	  	           nfreqview->y()+p.fontMetrics().ascent()
			           +p.fontMetrics().height(), i18n("Note:") );

     p.setPen( qRgb(0xFF,0xFF,0xFF) );
	   p.drawText( sampfreq_input->x()-p.fontMetrics().width(i18n("Sample Freq.:"))-5+1,
	  	           sampfreq_input->y()+p.fontMetrics().ascent()+1
			           +p.fontMetrics().height()-3, i18n("Sample Freq.:") );
     p.setPen( qRgb(0,0,0) );
	   p.drawText( sampfreq_input->x()-p.fontMetrics().width(i18n("Sample Freq.:"))-5,
	  	           sampfreq_input->y()+p.fontMetrics().ascent()
			           +p.fontMetrics().height()-3, i18n("Sample Freq.:") );

     p.setPen( qRgb(0xFF,0xFF,0xFF) );
	   p.drawText( sampnr_input->x()-p.fontMetrics().width(i18n("Sample #:"))-5+1,
	  	           sampnr_input->y()+p.fontMetrics().ascent()+1
			           +p.fontMetrics().height()-3, i18n("Sample #:") );
     p.setPen( qRgb(0,0,0) );
	   p.drawText( sampnr_input->x()-p.fontMetrics().width(i18n("Sample #:"))-5,
	  	           sampnr_input->y()+p.fontMetrics().ascent()
			           +p.fontMetrics().height()-3, i18n("Sample #:") );

     p.setPen( qRgb(0xFF,0xFF,0xFF) );
	   p.drawText( notch_input->x()-p.fontMetrics().width(i18n("Filter:"))-5+1,
	  	           notch_input->y()+p.fontMetrics().ascent()+1
			           +p.fontMetrics().height()-3, i18n("Filter:") );
     p.setPen( qRgb(0,0,0) );
	   p.drawText( notch_input->x()-p.fontMetrics().width(i18n("Filter:"))-5,
	  	           notch_input->y()+p.fontMetrics().ascent()
			           +p.fontMetrics().height()-3, i18n("Filter:") );
   p.end();

//   oszi->repaint();
   logview->update();
}

// This is needed to prevent the logview from getting drawn over the oszi initially
void MainWidget::showEvent( QShowEvent * )
{
   resizeEvent(0);
}

void MainWidget::resizeEvent( QResizeEvent * )
{
   int fwidth;

   // try to calculate width of freqview... quite ugly :)
   fwidth = (width() - sampfreq_input->x()) + QFontMetrics(QFont(QApplication::font().family(), 8)).width(i18n("Sample Freq.:")) + 10;

   oszi->setGeometry( 0, 0, width()-fwidth, height()-100 );
   if ( oszi->isVisible() )
   {
	logview->setGeometry( 0, oszi->height(), width(), 100 );
   } else {
	logview->setGeometry( 0, 0, width()-fwidth, 100 );
   }
//   freqview->setGeometry( width()-fwidth, ((oszi->isVisible())?oszi->height():0)+((logview->isVisible())?logview->height():0), 150, 30 );
//   nfreqview->setGeometry( width()-fwidth, ((oszi->isVisible())?oszi->height():0)+((logview->isVisible())?logview->height():0)+40, 150, 30 );
   freqview->setGeometry( width()-110, 3, 100, 25 );
   nfreqview->setGeometry( width()-110, 35, 100, 25 );
   sampfreq_input -> setGeometry( width()-80,65,70,20 );
   sampnr_input -> setGeometry( sampfreq_input->x(),
                                sampfreq_input->y()+
                                sampfreq_input->height()+5,  70,20 );
   notch_input -> setGeometry( sampnr_input->x()+20,
                               sampnr_input->y()+
                               sampnr_input->height()+5,  50,20 );
}


MainWidget::~MainWidget()
{
//   delete oszi;
   stop_audio();
}
