/***************************************************************************
                          soundinput.cpp  -  description
                             -------------------
    begin                : di apr 29 2003
    copyright            : (C) 2003 by Reinier Lamers
    email                : tux_rocker@planet.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

 // 15-08-2005: Added JACK support, made OSS and ALSA drivers do non-blocking opens - Reinier Lamers
 // 22-08-2003: Added ALSA support, and included config.h and added #ifdefs for audio system selection at compile time - Reinier Lamers

 #include <config.h>
 
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
 #include <exception>
 #include <stdexcept>

 #ifdef USE_OSS
 /* includes for OSS input driver */ 
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <sys/ioctl.h>
 #include <linux/soundcard.h>
 #endif

 #ifdef USE_ARTS
/* includes for Arts input driver */
/* I use artsc of all things because:
 *  - it doesn't automatically start an aRts server when there isn't one running already
 *  - it's simple
 *  - it actually works, which I doubt in the case of KAudioRecordStream
 */
 #include <artsc/artsc.h>
 #endif

 #ifdef USE_ALSA
 /* include for ALSA */
 #include <alsa/asoundlib.h>
 #include <asm/errno.h> /* for EAGAIN */
 #endif

 #ifdef USE_JACK
 // include files for JACK via bio2jack
 #include <bio2jack.h>
 #include <time.h>
 #endif

 #include "soundinput.h"

/*
 * Autodetects which sound system should be used and returns a corresponding
 * SoundInput. Returns 0 if no valid sound system was found.
 */
SoundInput *SoundInput::autodetect( const char *alsa_devicename, const char *devicename, int &blksize, int &sampfreq, double &sampfreq_exact )
{
  SoundInput *soundinput = 0;

#ifdef USE_JACK
  /* first try JACK. It may be that JACK is running, holding the audio device,
     and that aRts is running, having released the device. Then we must connect
     to JACK */
  try {
    soundinput = new JACKSoundInput( blksize, sampfreq, sampfreq_exact );
  } catch (SoundInputException) {
#endif
#ifdef USE_ARTS
    /* Then try aRts before ALSA and OSS, because aRts may be running having 
       released the device. We do not want to take the device ourselves then. */
    try {
        soundinput = new ArtsSoundInput(blksize, sampfreq, sampfreq_exact);
	} catch (SoundInputException) {
#endif
#ifdef USE_ALSA
	    /* Then try ALSA, because we do not want to try OSS and find ALSA's OSS emulation */
	    try {
	      soundinput = new ALSASoundInput( alsa_devicename, blksize, sampfreq, sampfreq_exact );
	    } catch (SoundInputException) {
#endif
#ifdef USE_OSS
	      /* If no aRts or ALSA present, try good old OSS */
	      try {
	        soundinput = new OSSSoundInput( devicename, blksize, sampfreq, sampfreq_exact );
	      } catch (SoundInputException) {
	      }
#endif
#ifdef USE_ALSA
	    }
#endif
#ifdef USE_ARTS
        }
#endif
#ifdef USE_JACK
  }
#endif
  return soundinput;
}
 
/* this solves some linker problems. Really needed? Is it time to go actually learn C++? */
SoundInput::~SoundInput( void )
{
}

#ifdef USE_OSS 
 /*
  * This function is mostly copy-n-pasted from the original MainWidget::init_audio(),
  * (c) Florian Berger
  */
 OSSSoundInput::OSSSoundInput( const char *dsp_devicename, int& blksize, int& sampfreq, double& sampfreq_exact )
 {
   printf("initializing audio at %s\n",dsp_devicename);

   // first, open the device non-blocking so that k3guitune won't hang when
   // the device is already taken
   fd = open(dsp_devicename, O_RDONLY | O_NONBLOCK); 
   if (fd == -1) {
      perror(dsp_devicename);
      throw SoundInputException();
   }
   fcntl(fd,F_SETFL, 0); // disable the non-blocking
   fcntl(fd,F_SETFD,FD_CLOEXEC);

//   ioctl(audio, SNDCTL_DSP_RESET, 0);
   if( strcmp(dsp_devicename,"/dev/stdin")==0 ){

      printf("reading data fronm stdin\n");
   		blksize=32;
      printf("  blocksize=%d\n",blksize);
      printf("  sampfreq=%d\n",sampfreq);
      sampfreq_exact = sampfreq;
   }else{

      ioctl(fd, SNDCTL_DSP_SETDUPLEX, 0);

      {int caps;
          ioctl( fd, SNDCTL_DSP_GETCAPS, &caps );
          printf("OSS-Version %d\n", caps & DSP_CAP_REVISION );
       		printf("  DUPLEX   = %X\n",caps & DSP_CAP_DUPLEX   );
          printf("  REALTIME = %X\n",caps & DSP_CAP_REALTIME );
          printf("  BATCH    = %X\n",caps & DSP_CAP_BATCH    );
          printf("  COPROC   = %X\n",caps & DSP_CAP_COPROC   );
          printf("  TRIGGER  = %X\n",caps & DSP_CAP_TRIGGER  );
          printf("  MMAP     = %X\n",caps & DSP_CAP_MMAP     );
   		}

   		blksize=8;  // 2^8 = 256
  		ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &blksize);

      ioctl(fd, SNDCTL_DSP_GETBLKSIZE, &blksize);
      printf("blocksize=%d\n",blksize);

      ioctl(fd, SNDCTL_DSP_SYNC, NULL);
      int a_sampsize = AFMT_S16_LE;
      ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &a_sampsize);
      int a_stereo = 0;
      ioctl(fd, SNDCTL_DSP_STEREO, &a_stereo);

      int a_speed = sampfreq;
      printf("sampfreq=%d\n",sampfreq);
      ioctl(fd, SNDCTL_DSP_SPEED, &a_speed);
      ioctl(fd, SOUND_PCM_READ_RATE, &sampfreq);
      printf("sampfreq=%d\n",sampfreq);
      sampfreq_exact = sampfreq;
   }
}

OSSSoundInput::~OSSSoundInput( void )
{
  close( fd );
}

int OSSSoundInput::getData( short int *buf, int blksize )
{
  return (int)read( fd, (void *)buf, blksize ) / 2;
}
#endif /* USE_OSS */

#ifdef USE_ARTS
ArtsSoundInput::ArtsSoundInput( int &blksize, int &sampfreq, double &sampfreq_exact )
{
  int errcode = arts_init();
  if (errcode)
  {
    fprintf( stderr, "Couldn't initialize Arts: %s\n", arts_error_text(errcode) );
    throw SoundInputException();
  }
  printf( "creating arts record stream, sampfreq=%i, 16 bits, 1 channel\n", sampfreq );
  stream = arts_record_stream( sampfreq, 16, 1, "k3guitune" );
  arts_stream_set(stream, ARTS_P_BLOCKING, 1);

  sampfreq_exact = sampfreq;

#if 0
  dispatcher = new KArtsDispatcher();
  server = new KArtsServer();
  stream = new KAudioRecordStream( server, "K3GuiTune::In", server );
  stream->usePolling( true );
  stream->setBlockingIO( false );
  
  printf( "Starting KAudioRecordStream with sampfreq=%i, 16 bits, 1 channel\n", sampfreq );
  stream->start( sampfreq, 16, 1 );
  sampfreq_exact = sampfreq;
  fprintf(stderr, "stream->running is %d\n", (int)stream->running());
#endif  
}

ArtsSoundInput::~ArtsSoundInput( void )
{
 arts_close_stream( stream );
 arts_free();
#if 0
  /*this should delete stream too */
  delete server;
  delete dispatcher;
#endif
}

int ArtsSoundInput::getData( short int *buf, int blksize )
{
  return arts_read( stream, (void *)buf, blksize ) / 2;
#if 0
  fprintf(stderr, "ArtsSoundInput::getData called\n");
  //stream->flush();
  //fprintf(stderr, "ArtsSoundInput::getData: Buffer geflusht\n");
  int ret = stream->read( (char *)buf, blksize );
  if (ret != 0) {
  	fprintf(stderr, "ArtsSoundInput::getData returning %d\n", ret);
  }
  return ret;
#endif
}
#endif /* USE_ARTS */

#ifdef USE_ALSA
ALSASoundInput::ALSASoundInput( const char *devicename, int &blksize, int &sampfreq, double &sampfreq_exact )
{
  snd_pcm_hw_params_t *hwparams;
  int dir;        /* used to store direction in which sample rate has been modified. quite useless */
  int err;

  snd_pcm_hw_params_malloc( &hwparams );

  /* create a PCM handle that is pointed to by this->pcm_handle
     such a handle will be what we operate on further */
  if ((err = snd_pcm_open(&pcm_handle, devicename, SND_PCM_STREAM_CAPTURE, SND_PCM_NONBLOCK)) < 0)
  {
    fprintf( stderr, "Couldn't open ALSA audio device %s: %s\n", devicename, snd_strerror(err) );
    snd_pcm_hw_params_free( hwparams );
    throw SoundInputException();
  }

  /* Set the handle to be blocking, so that we won't waste CPU cycles waiting for input */
  if ((err = snd_pcm_nonblock(pcm_handle, 0)) < 0)
  {
    fprintf(stderr, "Couldn't set blocking I/O for ALSA device %s: %s\n", devicename, snd_strerror(err));
    snd_pcm_hw_params_free(hwparams);
    snd_pcm_close(pcm_handle);
    throw SoundInputException();
  }

  /* Get the sound card configuration in hwparams */
  if ((err = snd_pcm_hw_params_any(pcm_handle, hwparams)) < 0)
  {
    fprintf( stderr, "Couldn't get sound card configuration for ALSA device %s: %s\n", devicename, snd_strerror(err) );
    snd_pcm_hw_params_free( hwparams );
    snd_pcm_close( pcm_handle );
    throw SoundInputException();
  }

  /* Set everything we have to set */
  if ((err = snd_pcm_hw_params_set_access(pcm_handle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0)
  {
    fprintf( stderr, "Couldn't set access for ALSA device %s: %s\n", devicename, snd_strerror(err) );
    snd_pcm_hw_params_free( hwparams );
    snd_pcm_close( pcm_handle );
    throw SoundInputException();
  }

  if ((err = snd_pcm_hw_params_set_format(pcm_handle, hwparams, SND_PCM_FORMAT_S16_LE)) < 0)
  {
    fprintf( stderr, "Couldn't set format for ALSA device %s: %s\n", devicename, snd_strerror(err) );
    snd_pcm_hw_params_free( hwparams );
    snd_pcm_close( pcm_handle );
    throw SoundInputException();
  }

  if ((err = snd_pcm_hw_params_set_rate_near(pcm_handle, hwparams, (unsigned *)&sampfreq, &dir)) < 0)
  {
    fprintf( stderr, "Couldn't set sample frequency for ALSA device %s: %s\n", devicename, snd_strerror(err) );
    snd_pcm_hw_params_free( hwparams );
    snd_pcm_close( pcm_handle );
    throw SoundInputException();
  }
  sampfreq_exact = sampfreq;

  if ((err = snd_pcm_hw_params_set_channels(pcm_handle, hwparams, 1)) < 0)
  {
    fprintf( stderr, "Couldn't set number of channels for ALSA device %s: %s\n", devicename, snd_strerror(err) );
    snd_pcm_hw_params_free( hwparams );
    snd_pcm_close( pcm_handle );
    throw SoundInputException();
  }

#if 0
  /* Don't really know what this function is good for, and it isn't used in aplay / arecord */
  if (snd_pcm_hw_params_set_periods(pcm_handle, hwparams, 2, 0) < 0)
  {
    fprintf( stderr, "Couldn't set number of periods for ALSA device %s\n", devicename );
    throw SoundInputException();
  }
#endif

#if 0 /* Frequently doesn't work */
  /* Use blksize / 2, because ALSA wants the size in (in our case, 16-bit) fragments instead of bytes */
  if ((err = snd_pcm_hw_params_set_buffer_size(pcm_handle, hwparams, blksize / 2)) < 0)
  {
    fprintf( stderr, "Couldn't set buffer size for ALSA device %s: %s\n", devicename, snd_strerror(err) );
    snd_pcm_hw_params_free( hwparams );
    snd_pcm_close( pcm_handle );
    throw SoundInputException();
  }
#endif

  /* Apply all these idiotic settings */
  if ((err = snd_pcm_hw_params(pcm_handle, hwparams)) < 0)
  {
    fprintf( stderr, "Couldn't set parameters for ALSA device %s: %s\n", devicename, snd_strerror(err) );
    snd_pcm_hw_params_free( hwparams );
    snd_pcm_close( pcm_handle );
    throw SoundInputException();
  }

  snd_pcm_hw_params_free( hwparams );
  printf( "ALSA audio initialized succesfully\n" );
}


ALSASoundInput::~ALSASoundInput( void )
{
  snd_pcm_close( pcm_handle );
}

int ALSASoundInput::getData( short int *buf, int blksize )
{
  int numblocks;

  numblocks = snd_pcm_readi( pcm_handle, (void *)buf, blksize / 2 );
  if (numblocks < 0)
  {
    /* I don't know if there is a more elegant solution to this. This works and it is badly needed */
    if (-numblocks != EAGAIN)
    {
      fprintf( stderr, "Error reading sound: %s\n", snd_strerror(numblocks) );
      snd_pcm_prepare( pcm_handle );
    }
    numblocks = 0;
  }
  return numblocks;
}
#endif /* USE_ALSA */

#ifdef USE_JACK
JACKSoundInput::JACKSoundInput(int &blksize, int &sampfreq, double &sampfreq_exact)
{
	int ret;
	unsigned long srateUL = (unsigned long)sampfreq;
	
	printf("Trying to set up JACK sound input\n");

	JACK_Init();
	JACK_SetClientName("k3guitune");

	/* open a device with 1 input channel and 0 output channels */
	if ((ret = JACK_OpenEx(&deviceID, 16, &srateUL, 1, 0, NULL, 0, 0)) != ERR_SUCCESS)
	{
		fprintf(stderr, "Error opening bio2jack device, error code is %d\n", ret);
		throw SoundInputException();
	}
	sampfreq = srateUL; /* save sample freq for our calller */
	sampfreq_exact = (double)sampfreq;
	printf("JACK sound input set up with sampfreq %f\n", sampfreq_exact);
}

JACKSoundInput::~JACKSoundInput(void)
{
	int ret;
	
	if((ret = JACK_Close(deviceID)) != ERR_SUCCESS)
	{
		fprintf(stderr, "Error closing bio2jack device: error code is %d\n", ret);
	}
}

/* In this function, we have to emulate the blocking ourselves because bio2jack
 * doesn't block. Limit the loop to 2 reads because otherwise UI responsivity
 * truly sucks. */
int JACKSoundInput::getData(short int *buf, int blksize)
{
	int bytes_read = 0;
	int rate = JACK_GetSampleRate(deviceID);
	int iterations = 0;
	struct timespec ts;
	
	while (bytes_read < blksize)
	{
		bytes_read += JACK_Read(deviceID, ((unsigned char *)buf + bytes_read), blksize - bytes_read);

		if (iterations >= 1) break;

		/* Calling nanosleep with no sleep time seems the only way to emulate blocking
		 * blocking succesfully, that is, keeping the UI responsive and keeping CPU
		 * usage low. Must have something to do with the timer system in the Linux
		 * kernel. The odd thing is that calling usleep or sleep with 0 as argument
		 * does generate way too much CPU usage. So does calling sched_yield().*/
		ts.tv_sec = 0;
		ts.tv_nsec = 0;
		nanosleep(&ts, NULL);

		iterations++;
	}

	return bytes_read / 2;
}
#endif /* USE_JACK */
